@@ -1,36 +1,49 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * e-tree-selection-model.c: a Selection Model
+ * e-tree-selection-model.c
+ * Copyright 2000, 2001, Ximian, Inc.
  *
- * Author:
- *   Christopher James Lahey <clahey@ximian.com>
+ * Authors:
+ *   Chris Lahey <clahey@ximian.com>
  *
- * (C) 2000, 2001 Ximian, Inc.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License, version 2, as published by the Free Software Foundation.
+ *
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
+ *
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
  */
+
 #include <config.h>
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtksignal.h>
 #include "e-tree-selection-model.h"
 #include <gal/util/e-bit-array.h>
 #include <gal/util/e-sorter.h>
+#include <gal/util/e-i18n.h>
 #include <gal/util/e-util.h>
 #include <gdk/gdkkeysyms.h>
 #include <gal/e-table/e-tree-sorted.h>
 #include <gal/e-table/e-tree-table-adapter.h>
 
-#define ETSM_CLASS(e) ((ETreeSelectionModelClass *)((GtkObject *)e)->klass)
-
 #define PARENT_TYPE e_selection_model_get_type ()
 
 static ESelectionModelClass *parent_class;
 
 enum {
-	ARG_0,
-	ARG_CURSOR_ROW,
-	ARG_CURSOR_COL,
-	ARG_MODEL,
-	ARG_ETTA,
-	ARG_ETS,
+	PROP_0,
+	PROP_CURSOR_ROW,
+	PROP_CURSOR_COL,
+	PROP_MODEL,
+	PROP_ETTA,
+	PROP_ETS
 };
 
 typedef struct ETreeSelectionModelNode {
@@ -56,19 +69,27 @@ struct ETreeSelectionModelPriv {
 
 	char *cursor_save_id;
 
+	int          frozen_count;
+
 	int          tree_model_pre_change_id;
+	int          tree_model_no_change_id;
 	int          tree_model_node_changed_id;
 	int          tree_model_node_data_changed_id;
 	int          tree_model_node_col_changed_id;
 	int          tree_model_node_inserted_id;
 	int          tree_model_node_removed_id;
+	int          tree_model_node_deleted_id;
 
-	int          sorted_model_pre_change_id;
-	int          sorted_model_node_changed_id;
-	int          sorted_model_node_data_changed_id;
-	int          sorted_model_node_col_changed_id;
-	int          sorted_model_node_inserted_id;
-	int          sorted_model_node_removed_id;
+	int          sorted_model_node_resorted_id;
+
+	/* Anything other than -1 means that the selection is a single
+	 * row.  This being -1 does not impart any information. */
+	int          selected_row;
+	/* Anything other than -1 means that the selection is a all
+	 * rows between selection_start_path and cursor_path where
+	 * selected_range_end is the rwo number of cursor_path.  This
+	 * being -1 does not impart any information. */
+	int          selected_range_end;
 };
 
 /* ETreeSelectionModelNode helpers */
@@ -106,9 +127,9 @@ e_tree_selection_model_node_free(ETreeSelectionModelNode *node)
 	int i;
 
 	if (node->all_children_selected_array)
-		gtk_object_unref(GTK_OBJECT(node->all_children_selected_array));
+		g_object_unref(node->all_children_selected_array);
 	if (node->any_children_selected_array)
-		gtk_object_unref(GTK_OBJECT(node->any_children_selected_array));
+		g_object_unref(node->any_children_selected_array);
 
 	for (i = 0; i < node->num_children; i++)
 		if (node->children[i])
@@ -196,7 +217,7 @@ etsm_find_node_unless_equals (ETreeSelectionModel *etsm,
 
 			return selection_node->children[position];
 		} else
-			  return NULL;
+			return NULL;
 	} else {
 		if (!etsm->priv->root)
 			etsm->priv->root = e_tree_selection_model_node_new();
@@ -330,8 +351,52 @@ update_parents (ETreeSelectionModel *etsm, ETreePath path)
 }
 
 
+static void
+etsm_change_one_path(ETreeSelectionModel *etsm, ETreePath path, gboolean grow)
+{
+	ETreeSelectionModelNode *node;
+
+	if (!path)
+		return;
+
+	path = e_tree_sorted_model_to_view_path(etsm->priv->ets, path);
+
+	if (!path)
+		return;
+
+	node = etsm_find_node_unless_equals (etsm, path, grow);
+
+	if (node) {
+		node->selected = grow;
+		update_parents(etsm, path);
+	}
+}
+
+static void
+etsm_real_select_single_path (ETreeSelectionModel *etsm, ETreePath path)
+{
+	etsm_real_clear (etsm);
+	etsm_change_one_path(etsm, path, TRUE);
+	etsm->priv->selection_start_path = path;
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
+}
+
+
 /* Signal handlers */
 
+static void
+etsm_freeze (ETreeSelectionModel *etsm)
+{
+	etsm->priv->frozen_count ++;
+}
+
+static void
+etsm_unfreeze (ETreeSelectionModel *etsm)
+{
+	etsm->priv->frozen_count --;
+}
+
 static void
 etsm_pre_change (ETreeModel *etm, ETreeSelectionModel *etsm)
 {
@@ -343,14 +408,23 @@ etsm_pre_change (ETreeModel *etm, ETreeSelectionModel *etsm)
 	    etsm->priv->cursor_path) {
 		etsm->priv->cursor_save_id = e_tree_model_get_save_id (etm, etsm->priv->cursor_path);
 	}
+
+	etsm_freeze (etsm);
 }
 
 static void
-etsm_node_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
+etsm_no_change (ETreeModel *etm, ETreeSelectionModel *etsm)
+{
+	etsm_unfreeze (etsm);
+}
+
+static void
+clear_tree (ETreeSelectionModel *etsm, ETreeModel *etm)
 {
 	int cursor_row;
 
 	etsm_real_clear (etsm);
+	etsm->priv->cursor_path = NULL;
 
 	if (e_tree_model_has_get_node_by_id (etm) && etsm->priv->cursor_save_id) {
 		ETreePath cursor_path;
@@ -360,10 +434,12 @@ etsm_node_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
 		if (cursor_path != NULL && etsm->priv->cursor_col == -1)
 			etsm->priv->cursor_col = 0;
 
-		e_tree_selection_model_select_single_path(etsm, cursor_path);
+		etsm_real_select_single_path(etsm, cursor_path);
 	}
 
 	cursor_row = etsm_cursor_row_real (etsm);
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 	if (cursor_row != -1)
 		e_selection_model_cursor_changed(E_SELECTION_MODEL(etsm), cursor_row, etsm->priv->cursor_col);
@@ -376,11 +452,19 @@ etsm_node_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
 	etsm->priv->cursor_save_id = NULL;
 }
 
+static void
+etsm_node_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
+{
+	clear_tree (etsm, etm);
+	etsm_unfreeze (etsm);
+}
+
 static void
 etsm_node_data_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
 {
 	g_free (etsm->priv->cursor_save_id);
 	etsm->priv->cursor_save_id = NULL;
+	etsm_unfreeze (etsm);
 }
 
 static void
@@ -388,12 +472,14 @@ etsm_node_col_changed (ETreeModel *etm, ETreePath node, int col, ETreeSelectionM
 {
 	g_free (etsm->priv->cursor_save_id);
 	etsm->priv->cursor_save_id = NULL;
+	etsm_unfreeze (etsm);
 }
 
 static void
 etsm_node_inserted (ETreeModel *etm, ETreePath parent, ETreePath child, ETreeSelectionModel *etsm)
 {
-	etsm_node_changed(etm, parent, etsm);
+	clear_tree (etsm, etm);
+	etsm_unfreeze (etsm);
 #if 0
 	ETreeSelectionModelNode *node;
 	ETreePath path;
@@ -415,20 +501,23 @@ etsm_node_inserted (ETreeModel *etm, ETreePath parent, ETreePath child, ETreeSel
 static void
 etsm_node_removed (ETreeModel *etm, ETreePath parent, ETreePath child, int old_position, ETreeSelectionModel *etsm)
 {
-	etsm_node_changed(etm, parent, etsm);
 }
 
-
-
 static void
-etsm_sorted_pre_change (ETreeModel *etm, ETreeSelectionModel *etsm)
+etsm_node_deleted (ETreeModel *etm, ETreePath child, ETreeSelectionModel *etsm)
 {
+	clear_tree (etsm, etm);
+	etsm_unfreeze (etsm);
 }
 
+
 static void
-etsm_sorted_node_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
+etsm_sorted_node_resorted (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
 {
 	int cursor_row = etsm_cursor_row_real (etsm);
+
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 
 	if (cursor_row != -1)
@@ -437,29 +526,6 @@ etsm_sorted_node_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *
 		e_selection_model_cursor_changed(E_SELECTION_MODEL(etsm), -1, -1);
 }
 
-static void
-etsm_sorted_node_data_changed (ETreeModel *etm, ETreePath node, ETreeSelectionModel *etsm)
-{
-}
-
-static void
-etsm_sorted_node_col_changed (ETreeModel *etm, ETreePath node, int col, ETreeSelectionModel *etsm)
-{
-}
-
-static void
-etsm_sorted_node_inserted (ETreeModel *etm, ETreePath parent, ETreePath child, ETreeSelectionModel *etsm)
-{
-	etsm_sorted_node_changed(etm, parent, etsm);
-}
-
-static void
-etsm_sorted_node_removed (ETreeModel *etm, ETreePath parent, ETreePath child, int old_position, ETreeSelectionModel *etsm)
-{
-	etsm_sorted_node_changed(etm, parent, etsm);
-}
-
-
 static void
 add_model(ETreeSelectionModel *etsm, ETreeModel *model)
 {
@@ -470,19 +536,23 @@ add_model(ETreeSelectionModel *etsm, ETreeModel *model)
 	if (!priv->model)
 		return;
 
-	gtk_object_ref(GTK_OBJECT(priv->model));
-	priv->tree_model_pre_change_id        = gtk_signal_connect (GTK_OBJECT (priv->model), "pre_change",
-								    GTK_SIGNAL_FUNC (etsm_pre_change), etsm);
-	priv->tree_model_node_changed_id      = gtk_signal_connect (GTK_OBJECT (priv->model), "node_changed",
-								    GTK_SIGNAL_FUNC (etsm_node_changed), etsm);
-	priv->tree_model_node_data_changed_id = gtk_signal_connect (GTK_OBJECT (priv->model), "node_data_changed",
-								    GTK_SIGNAL_FUNC (etsm_node_data_changed), etsm);
-	priv->tree_model_node_col_changed_id  = gtk_signal_connect (GTK_OBJECT (priv->model), "node_col_changed",
-								    GTK_SIGNAL_FUNC (etsm_node_col_changed), etsm);
-	priv->tree_model_node_inserted_id     = gtk_signal_connect (GTK_OBJECT (priv->model), "node_inserted",
-								    GTK_SIGNAL_FUNC (etsm_node_inserted), etsm);
-	priv->tree_model_node_removed_id      = gtk_signal_connect (GTK_OBJECT (priv->model), "node_removed",
-								    GTK_SIGNAL_FUNC (etsm_node_removed), etsm);
+	g_object_ref(priv->model);
+	priv->tree_model_pre_change_id        = g_signal_connect_after (G_OBJECT (priv->model), "pre_change",
+									G_CALLBACK (etsm_pre_change), etsm);
+	priv->tree_model_no_change_id         = g_signal_connect_after (G_OBJECT (priv->model), "no_change",
+									G_CALLBACK (etsm_no_change), etsm);
+	priv->tree_model_node_changed_id      = g_signal_connect_after (G_OBJECT (priv->model), "node_changed",
+									G_CALLBACK (etsm_node_changed), etsm);
+	priv->tree_model_node_data_changed_id = g_signal_connect_after (G_OBJECT (priv->model), "node_data_changed",
+									G_CALLBACK (etsm_node_data_changed), etsm);
+	priv->tree_model_node_col_changed_id  = g_signal_connect_after (G_OBJECT (priv->model), "node_col_changed",
+									G_CALLBACK (etsm_node_col_changed), etsm);
+	priv->tree_model_node_inserted_id     = g_signal_connect_after (G_OBJECT (priv->model), "node_inserted",
+									G_CALLBACK (etsm_node_inserted), etsm);
+	priv->tree_model_node_removed_id      = g_signal_connect_after (G_OBJECT (priv->model), "node_removed",
+									G_CALLBACK (etsm_node_removed), etsm);
+	priv->tree_model_node_deleted_id      = g_signal_connect_after (G_OBJECT (priv->model), "node_deleted",
+									G_CALLBACK (etsm_node_deleted), etsm);
 }
 
 static void
@@ -493,28 +563,34 @@ drop_model(ETreeSelectionModel *etsm)
 	if (!priv->model)
 		return;
 
-	gtk_signal_disconnect (GTK_OBJECT (priv->model),
-			       priv->tree_model_pre_change_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->model),
-			       priv->tree_model_node_changed_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->model),
-			       priv->tree_model_node_data_changed_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->model),
-			       priv->tree_model_node_col_changed_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->model),
-			       priv->tree_model_node_inserted_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->model),
-			       priv->tree_model_node_removed_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_pre_change_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_no_change_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_node_changed_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_node_data_changed_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_node_col_changed_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_node_inserted_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_node_removed_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->model),
+			             priv->tree_model_node_deleted_id);
 
-	gtk_object_unref (GTK_OBJECT (priv->model));
+	g_object_unref (priv->model);
 	priv->model = NULL;
 
 	priv->tree_model_pre_change_id = 0;
+	priv->tree_model_no_change_id = 0;
 	priv->tree_model_node_changed_id = 0;
 	priv->tree_model_node_data_changed_id = 0;
 	priv->tree_model_node_col_changed_id = 0;
 	priv->tree_model_node_inserted_id = 0;
 	priv->tree_model_node_removed_id = 0;
+	priv->tree_model_node_deleted_id = 0;
 }
 
 
@@ -528,19 +604,9 @@ add_ets(ETreeSelectionModel *etsm, ETreeSorted *ets)
 	if (!priv->ets)
 		return;
 
-	gtk_object_ref(GTK_OBJECT(priv->ets));
-	priv->sorted_model_pre_change_id        = gtk_signal_connect (GTK_OBJECT (priv->ets), "pre_change",
-								      GTK_SIGNAL_FUNC (etsm_sorted_pre_change), etsm);
-	priv->sorted_model_node_changed_id      = gtk_signal_connect (GTK_OBJECT (priv->ets), "node_changed",
-								    GTK_SIGNAL_FUNC (etsm_sorted_node_changed), etsm);
-	priv->sorted_model_node_data_changed_id = gtk_signal_connect (GTK_OBJECT (priv->ets), "node_data_changed",
-								    GTK_SIGNAL_FUNC (etsm_sorted_node_data_changed), etsm);
-	priv->sorted_model_node_col_changed_id  = gtk_signal_connect (GTK_OBJECT (priv->ets), "node_col_changed",
-								    GTK_SIGNAL_FUNC (etsm_sorted_node_col_changed), etsm);
-	priv->sorted_model_node_inserted_id     = gtk_signal_connect (GTK_OBJECT (priv->ets), "node_inserted",
-								    GTK_SIGNAL_FUNC (etsm_sorted_node_inserted), etsm);
-	priv->sorted_model_node_removed_id      = gtk_signal_connect (GTK_OBJECT (priv->ets), "node_removed",
-								    GTK_SIGNAL_FUNC (etsm_sorted_node_removed), etsm);
+	g_object_ref(priv->ets);
+	priv->sorted_model_node_resorted_id      = g_signal_connect (G_OBJECT (priv->ets), "node_resorted",
+								     G_CALLBACK (etsm_sorted_node_resorted), etsm);
 }
 
 static void
@@ -551,121 +617,113 @@ drop_ets(ETreeSelectionModel *etsm)
 	if (!priv->ets)
 		return;
 
-	gtk_signal_disconnect (GTK_OBJECT (priv->ets),
-			       priv->sorted_model_pre_change_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->ets),
-			       priv->sorted_model_node_changed_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->ets),
-			       priv->sorted_model_node_data_changed_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->ets),
-			       priv->sorted_model_node_col_changed_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->ets),
-			       priv->sorted_model_node_inserted_id);
-	gtk_signal_disconnect (GTK_OBJECT (priv->ets),
-			       priv->sorted_model_node_removed_id);
+	g_signal_handler_disconnect (G_OBJECT (priv->ets),
+			             priv->sorted_model_node_resorted_id);
 
-	gtk_object_unref (GTK_OBJECT (priv->ets));
+	g_object_unref (priv->ets);
 	priv->ets = NULL;
 
-	priv->sorted_model_pre_change_id = 0;
-	priv->sorted_model_node_changed_id = 0;
-	priv->sorted_model_node_data_changed_id = 0;
-	priv->sorted_model_node_col_changed_id = 0;
-	priv->sorted_model_node_inserted_id = 0;
-	priv->sorted_model_node_removed_id = 0;
+	priv->sorted_model_node_resorted_id = 0;
 }
 
 /* Virtual functions */
 static void
-etsm_destroy (GtkObject *object)
+etsm_dispose (GObject *object)
 {
 	ETreeSelectionModel *etsm;
 
 	etsm = E_TREE_SELECTION_MODEL (object);
 
-	etsm_real_clear (etsm);
-	drop_model(etsm);
-	drop_ets(etsm);
+	if (etsm->priv) {
+		etsm_real_clear (etsm);
+		etsm->priv->cursor_path = NULL;
 
-	g_free (etsm->priv->cursor_save_id);
+		drop_model(etsm);
+		drop_ets(etsm);
+		
+		g_free (etsm->priv->cursor_save_id);
+		etsm->priv->cursor_save_id = NULL;
 
-	g_free(etsm->priv);
-	etsm->priv = NULL;
+		g_free (etsm->priv);
+		etsm->priv = NULL;
+	}
+
+	if (G_OBJECT_CLASS (parent_class)->dispose)
+		(* G_OBJECT_CLASS (parent_class)->dispose) (object);
 }
 
 static void
-etsm_get_arg (GtkObject *o, GtkArg *arg, guint arg_id)
+etsm_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
 {
-	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL (o);
+	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL (object);
 
-	switch (arg_id){
-	case ARG_CURSOR_ROW:
-		GTK_VALUE_INT(*arg) = etsm_cursor_row_real(etsm);
+	switch (prop_id){
+	case PROP_CURSOR_ROW:
+		g_value_set_int (value, etsm_cursor_row_real(etsm));
 		break;
 
-	case ARG_CURSOR_COL:
-		GTK_VALUE_INT(*arg) = etsm->priv->cursor_col;
+	case PROP_CURSOR_COL:
+		g_value_set_int (value, etsm->priv->cursor_col);
 		break;
 
-	case ARG_MODEL:
-		GTK_VALUE_OBJECT(*arg) = (GtkObject *) etsm->priv->model;
+	case PROP_MODEL:
+		g_value_set_object (value, etsm->priv->model);
 		break;
 
-	case ARG_ETTA:
-		GTK_VALUE_OBJECT(*arg) = (GtkObject *) etsm->priv->etta;
+	case PROP_ETTA:
+		g_value_set_object (value, etsm->priv->etta);
 		break;
 
-	case ARG_ETS:
-		GTK_VALUE_OBJECT(*arg) = (GtkObject *) etsm->priv->ets;
+	case PROP_ETS:
+		g_value_set_object (value, etsm->priv->ets);
 		break;
 	}
 }
 
 static void
-etsm_set_arg (GtkObject *o, GtkArg *arg, guint arg_id)
+etsm_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
 {
-	ESelectionModel *esm = E_SELECTION_MODEL (o);
-	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL (o);
+	ESelectionModel *esm = E_SELECTION_MODEL (object);
+	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL (object);
 
-	switch (arg_id){
-	case ARG_CURSOR_ROW:
-		e_selection_model_do_something(esm, GTK_VALUE_INT(*arg), etsm->priv->cursor_col, 0);
+	switch (prop_id){
+	case PROP_CURSOR_ROW:
+		e_selection_model_do_something(esm, g_value_get_int (value), etsm->priv->cursor_col, 0);
 		break;
 
-	case ARG_CURSOR_COL:
-		e_selection_model_do_something(esm, etsm_cursor_row_real(etsm), GTK_VALUE_INT(*arg), 0);
+	case PROP_CURSOR_COL:
+		e_selection_model_do_something(esm, etsm_cursor_row_real(etsm), g_value_get_int(value), 0);
 		break;
 
-	case ARG_MODEL:
+	case PROP_MODEL:
 		drop_model(etsm);
-		add_model(etsm, (ETreeModel *) GTK_VALUE_OBJECT(*arg));
+		add_model(etsm, E_TREE_MODEL (g_value_get_object(value)));
 		break;
 
-	case ARG_ETTA:
-		etsm->priv->etta = (ETreeTableAdapter *) GTK_VALUE_OBJECT(*arg);
+	case PROP_ETTA:
+		etsm->priv->etta = E_TREE_TABLE_ADAPTER (g_value_get_object (value));
 		break;
 
-	case ARG_ETS:
+	case PROP_ETS:
 		drop_ets(etsm);
-		add_ets(etsm, (ETreeSorted *) GTK_VALUE_OBJECT(*arg));
+		add_ets(etsm, E_TREE_SORTED (g_value_get_object (value)));
 		break;
 	}
 }
 
 static ETreeSelectionModelNode *
-etsm_recurse_is_path_selected (ESelectionModel *selection,
+etsm_recurse_is_path_selected (ETreeSelectionModel *etsm,
 			       ETreePath path,
 			       gboolean *is_selected)
 {
 	ETreeSelectionModelNode *selection_node;
-	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
 	ETreeSorted *ets = etsm->priv->ets;
 	ETreePath parent;
 
 	parent = e_tree_model_node_get_parent(E_TREE_MODEL(ets), path);
 
 	if (parent) {
-		selection_node = etsm_recurse_is_path_selected (selection, parent, is_selected);
+		selection_node = etsm_recurse_is_path_selected (etsm, parent, is_selected);
 		if (selection_node) {
 			int position = e_tree_sorted_orig_position(ets, path);
 			if (position < 0 || position >= selection_node->num_children) {
@@ -705,6 +763,21 @@ etsm_recurse_is_path_selected (ESelectionModel *selection,
 	}
 }
 
+static gboolean
+etsm_is_path_selected (ETreeSelectionModel *etsm,
+		       ETreePath path)
+{
+	ETreeSelectionModelNode *selection_node;
+	gboolean ret_val;
+
+	selection_node = etsm_recurse_is_path_selected (etsm, path, &ret_val);
+
+	if (selection_node)
+		ret_val = selection_node->selected;
+
+	return ret_val;
+}
+
 /** 
  * e_selection_model_is_row_selected
  * @selection: #ESelectionModel to check
@@ -720,22 +793,13 @@ etsm_is_row_selected (ESelectionModel *selection,
 {
 	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
 	ETreePath path;
-	ETreeSelectionModelNode *selection_node;
-
-	gboolean ret_val;
 
 	g_return_val_if_fail(row < e_table_model_row_count(E_TABLE_MODEL(etsm->priv->etta)), FALSE);
 	g_return_val_if_fail(row >= 0, FALSE);
-	g_return_val_if_fail(selection != NULL, FALSE);
+	g_return_val_if_fail(etsm != NULL, FALSE);
 
 	path = e_tree_table_adapter_node_at_row(etsm->priv->etta, row);
-
-	selection_node = etsm_recurse_is_path_selected (selection, path, &ret_val);
-
-	if (selection_node)
-		ret_val = selection_node->selected;
-
-	return ret_val;
+	return etsm_is_path_selected (etsm, path);
 }
 
 
@@ -750,7 +814,7 @@ etsm_row_foreach_cb (ETreePath path, gpointer user_data)
 {
 	ModelAndCallback *mac = user_data;
 	int row = etsm_row_of_node(mac->etsm, path);
-	if (row != -1)
+	if (row >= 0)
 		mac->callback(row, mac->closure);
 }
 
@@ -793,11 +857,59 @@ etsm_clear(ESelectionModel *selection)
 
 	etsm->priv->cursor_path = NULL;
 	etsm->priv->cursor_col = -1;
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 	e_selection_model_cursor_changed(E_SELECTION_MODEL(etsm), -1, -1);
 }
 
-#if 0
+/* Standard functions */
+static void
+etsm_selected_count_all_recurse (ETreeSelectionModel *etsm,
+				 ETreePath path,
+				 int *count)
+{
+	ETreePath child;
+
+	(*count) ++;
+
+	child = e_tree_model_node_get_first_child(E_TREE_MODEL(etsm->priv->model), path);
+	for ( ; child; child = e_tree_model_node_get_next(E_TREE_MODEL(etsm->priv->model), child))
+		if (child)
+			etsm_selected_count_all_recurse (etsm, child, count);
+}
+
+static void
+etsm_selected_count_recurse (ETreeSelectionModel *etsm,
+			     ETreeSelectionModelNode *selection_node,
+			     ETreePath path,
+			     int *count)
+{
+	if (selection_node->all_children_selected) {
+		if (path)
+			etsm_selected_count_all_recurse(etsm, path, count);
+		return;
+	}
+
+	if (selection_node->selected) {
+		(*count) ++;
+	}
+
+	if (!selection_node->any_children_selected)
+		return;
+
+	if (selection_node->children) {
+		ETreePath child = e_tree_model_node_get_first_child(E_TREE_MODEL(etsm->priv->model), path);
+		int i;
+		for (i = 0; child && i < selection_node->num_children; i++, child = e_tree_model_node_get_next(E_TREE_MODEL(etsm->priv->model), child))
+			if (selection_node->all_children_selected_array && e_bit_array_value_at(selection_node->all_children_selected_array, i))
+				etsm_selected_count_all_recurse (etsm, child, count);
+			else if (selection_node->children[i])
+				etsm_selected_count_recurse (etsm, selection_node->children[i], child, count);
+				
+	}
+}
+
 /** 
  * e_selection_model_selected_count
  * @selection: #ESelectionModel to count
@@ -810,10 +922,17 @@ static gint
 etsm_selected_count (ESelectionModel *selection)
 {
 	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
-
-	return g_hash_table_size(etsm->priv->data);
+	int count = 0;
+	if (etsm->priv->root) {
+		ETreePath model_root;
+		model_root = e_tree_model_get_root(etsm->priv->model);
+		etsm_selected_count_recurse(etsm, etsm->priv->root, model_root, &count);
+		if (!e_tree_table_adapter_root_node_is_visible (etsm->priv->etta) && etsm_is_path_selected (etsm, e_tree_model_get_root(E_TREE_MODEL (etsm->priv->ets)))) {
+			count --;
+		}
+	}
+	return count;
 }
-#endif
 
 /** 
  * e_selection_model_select_all
@@ -826,6 +945,11 @@ static void
 etsm_select_all (ESelectionModel *selection)
 {
 	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
+	ETreePath root;
+
+	root = e_tree_model_get_root(E_TREE_MODEL(etsm->priv->ets));
+	if (root == NULL)
+		return;
 
 	etsm_real_clear (etsm);
 
@@ -834,17 +958,17 @@ etsm_select_all (ESelectionModel *selection)
 	etsm->priv->root->all_children_selected = TRUE;
 	etsm->priv->root->any_children_selected = TRUE;
 
-	e_tree_selection_model_node_fill_children(etsm, e_tree_model_get_root(E_TREE_MODEL(etsm->priv->ets)), etsm->priv->root);
-	etsm->priv->root->all_children_selected_array = e_bit_array_new(etsm->priv->root->num_children);
-	etsm->priv->root->any_children_selected_array = e_bit_array_new(etsm->priv->root->num_children);
-	e_bit_array_select_all(etsm->priv->root->all_children_selected_array);
-	e_bit_array_select_all(etsm->priv->root->any_children_selected_array);
+	e_tree_selection_model_node_fill_children(etsm, root, etsm->priv->root);
+	etsm->priv->root->all_children_selected_array = NULL;
+	etsm->priv->root->any_children_selected_array = NULL;
 
 	if (etsm->priv->cursor_col == -1)
 		etsm->priv->cursor_col = 0;
 	if (etsm->priv->cursor_path == NULL)
 		etsm->priv->cursor_path = etsm_node_at_row(etsm, 0);
 	etsm->priv->selection_start_path = etsm_node_at_row(etsm, 0);
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 	e_selection_model_cursor_changed(E_SELECTION_MODEL(etsm), etsm_cursor_row_real(etsm), etsm->priv->cursor_col);
 }
@@ -895,6 +1019,8 @@ etsm_invert_selection (ESelectionModel *selection)
 	etsm->priv->cursor_col = -1;
 	etsm->priv->cursor_path = NULL;
 	etsm->priv->selection_start_path = etsm_node_at_row(etsm, 0);
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 	e_selection_model_cursor_changed(E_SELECTION_MODEL(etsm), -1, -1);
 }
@@ -930,27 +1056,6 @@ etsm_change_one_row(ESelectionModel *selection, int row, gboolean grow)
 	}
 }
 
-static void
-etsm_change_one_path(ETreeSelectionModel *etsm, ETreePath path, gboolean grow)
-{
-	ETreeSelectionModelNode *node;
-
-	if (!path)
-		return;
-
-	path = e_tree_sorted_model_to_view_path(etsm->priv->ets, path);
-
-	if (!path)
-		return;
-
-	node = etsm_find_node_unless_equals (etsm, path, grow);
-
-	if (node) {
-		node->selected = grow;
-		update_parents(etsm, path);
-	}
-}
-
 static void
 etsm_change_cursor (ESelectionModel *selection, int row, int col)
 {
@@ -1003,12 +1108,22 @@ static void
 etsm_select_single_row (ESelectionModel *selection, int row)
 {
 	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
+	int selected_row = etsm->priv->selected_row;
 
 	etsm_real_clear (etsm);
 	etsm_change_one_row(selection, row, TRUE);
 	etsm->priv->selection_start_path = etsm_node_at_row(etsm, row);
 
-	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
+	etsm->priv->selected_row = row;
+	etsm->priv->selected_range_end = -1;
+	if (selected_row != -1) {
+		if (selected_row != etsm->priv->selected_row) {
+			e_selection_model_selection_row_changed(E_SELECTION_MODEL(etsm), selected_row);
+			e_selection_model_selection_row_changed(E_SELECTION_MODEL(etsm), row);
+		}
+	} else {
+		e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
+	}
 }
 
 static void
@@ -1020,13 +1135,15 @@ etsm_toggle_single_row (ESelectionModel *selection, int row)
 
 	etsm_change_one_row(selection, row, !etsm_is_row_selected(selection, row));
 
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 }
 
 static void
-etsm_move_selection_end (ESelectionModel *selection, int row)
+etsm_real_move_selection_end (ETreeSelectionModel *etsm, int row)
 {
-	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
+	ESelectionModel *selection = E_SELECTION_MODEL (etsm);
 	int old_start;
 	int old_end;
 	int new_start;
@@ -1056,16 +1173,38 @@ etsm_move_selection_end (ESelectionModel *selection, int row)
 		etsm_change_range(selection, old_end, new_end, TRUE);
 	if (new_end < old_end)
 		etsm_change_range(selection, new_end, old_end, FALSE);
-	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
+}
+
+static void
+etsm_move_selection_end (ESelectionModel *selection, int row)
+{
+	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
+
+	etsm_real_move_selection_end (etsm, row);
+	e_selection_model_selection_changed(E_SELECTION_MODEL(selection));
 }
 
 static void
 etsm_set_selection_end (ESelectionModel *selection, int row)
 {
 	ETreeSelectionModel *etsm = E_TREE_SELECTION_MODEL(selection);
-	e_tree_selection_model_select_single_path(etsm, etsm->priv->selection_start_path);
+	int selected_range_end = etsm->priv->selected_range_end;
+
+	etsm_real_select_single_path(etsm, etsm->priv->selection_start_path);
 	etsm->priv->cursor_path = etsm->priv->selection_start_path;
-	e_selection_model_move_selection_end(selection, row);
+	etsm_real_move_selection_end(etsm, row);
+	etsm->priv->selected_range_end = row;
+	if (selected_range_end != -1 && row != -1) {
+		if (selected_range_end == row - 1 ||
+		    selected_range_end == row + 1) {
+			e_selection_model_selection_row_changed(E_SELECTION_MODEL(etsm), selected_range_end);
+			e_selection_model_selection_row_changed(E_SELECTION_MODEL(etsm), row);
+			return;
+		}
+	}
+	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 }
 
 
@@ -1109,7 +1248,9 @@ etsm_foreach_recurse (ETreeSelectionModel *etsm,
 		ETreePath child = e_tree_model_node_get_first_child(E_TREE_MODEL(etsm->priv->model), path);
 		int i;
 		for (i = 0; i < selection_node->num_children; i++, child = e_tree_model_node_get_next(E_TREE_MODEL(etsm->priv->model), child))
-			if (selection_node->children[i])
+			if (selection_node->all_children_selected_array && e_bit_array_value_at(selection_node->all_children_selected_array, i))
+				etsm_foreach_all_recurse(etsm, child, callback, closure);
+			else if (selection_node->children[i])
 				etsm_foreach_recurse (etsm, selection_node->children[i], child, callback, closure);
 	}
 }
@@ -1129,9 +1270,18 @@ e_tree_selection_model_foreach   (ETreeSelectionModel *etsm,
 void
 e_tree_selection_model_select_single_path (ETreeSelectionModel *etsm, ETreePath path)
 {
-	etsm_real_clear (etsm);
+	etsm_real_select_single_path (etsm, path);
+
+	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
+}
+
+void
+e_tree_selection_model_add_to_selection (ETreeSelectionModel *etsm, ETreePath path)
+{
 	etsm_change_one_path(etsm, path, TRUE);
 	etsm->priv->selection_start_path = path;
+	etsm->priv->selected_row = -1;
+	etsm->priv->selected_range_end = -1;
 
 	e_selection_model_selection_changed(E_SELECTION_MODEL(etsm));
 }
@@ -1145,67 +1295,76 @@ e_tree_selection_model_change_cursor (ETreeSelectionModel *etsm, ETreePath path)
 
 	row = etsm_cursor_row_real(etsm);
 
+	E_SELECTION_MODEL (etsm)->old_selection = -1;
+
 	e_selection_model_cursor_changed(E_SELECTION_MODEL(etsm), row, etsm->priv->cursor_col);
 	e_selection_model_cursor_activated(E_SELECTION_MODEL(etsm), row, etsm->priv->cursor_col);
 }
 
+ETreePath
+e_tree_selection_model_get_cursor (ETreeSelectionModel *etsm)
+{
+	return etsm->priv->cursor_path;
+}
+
 
 static void
 e_tree_selection_model_init (ETreeSelectionModel *etsm)
 {
 	ETreeSelectionModelPriv *priv;
-	priv                                    = g_new(ETreeSelectionModelPriv, 1);
-	etsm->priv                              = priv;
+	priv                                  = g_new(ETreeSelectionModelPriv, 1);
+	etsm->priv                            = priv;
 
-	priv->etta                              = NULL;
-	priv->ets                               = NULL;
-	priv->model                             = NULL;
+	priv->etta                            = NULL;
+	priv->ets                             = NULL;
+	priv->model                           = NULL;
 
-	priv->root                              = NULL;
+	priv->root                            = NULL;
 
-	priv->cursor_path                       = NULL;
-	priv->cursor_col                        = -1;
-	priv->selection_start_path              = NULL;
+	priv->cursor_path                     = NULL;
+	priv->cursor_col                      = -1;
+	priv->selection_start_path            = NULL;
 
-	priv->cursor_save_id                    = NULL;
+	priv->cursor_save_id                  = NULL;
+
+	priv->frozen_count                    = 0;
+
+	priv->frozen_count                    = 0;
 
 
-	priv->tree_model_pre_change_id          = 0;
-	priv->tree_model_node_changed_id        = 0;
-	priv->tree_model_node_data_changed_id   = 0;
-	priv->tree_model_node_col_changed_id    = 0;
-	priv->tree_model_node_inserted_id       = 0;
-	priv->tree_model_node_removed_id        = 0;
+	priv->tree_model_pre_change_id        = 0;
+	priv->tree_model_no_change_id         = 0;
+	priv->tree_model_node_changed_id      = 0;
+	priv->tree_model_node_data_changed_id = 0;
+	priv->tree_model_node_col_changed_id  = 0;
+	priv->tree_model_node_inserted_id     = 0;
+	priv->tree_model_node_removed_id      = 0;
+	priv->tree_model_node_deleted_id      = 0;
 
-	priv->sorted_model_pre_change_id        = 0;
-	priv->sorted_model_node_changed_id      = 0;
-	priv->sorted_model_node_data_changed_id = 0;
-	priv->sorted_model_node_col_changed_id  = 0;
-	priv->sorted_model_node_inserted_id     = 0;
-	priv->sorted_model_node_removed_id      = 0;
+	priv->sorted_model_node_resorted_id   = 0;
+	priv->selected_row                    = -1;
+	priv->selected_range_end              = -1;
 }
 
 static void
 e_tree_selection_model_class_init (ETreeSelectionModelClass *klass)
 {
-	GtkObjectClass *object_class;
+	GObjectClass *object_class;
 	ESelectionModelClass *esm_class;
 
-	parent_class = gtk_type_class (e_selection_model_get_type ());
+	parent_class = g_type_class_ref (PARENT_TYPE);
 
-	object_class = GTK_OBJECT_CLASS(klass);
+	object_class = G_OBJECT_CLASS(klass);
 	esm_class = E_SELECTION_MODEL_CLASS(klass);
 
-	object_class->destroy = etsm_destroy;
-	object_class->get_arg = etsm_get_arg;
-	object_class->set_arg = etsm_set_arg;
+	object_class->dispose = etsm_dispose;
+	object_class->get_property = etsm_get_property;
+	object_class->set_property = etsm_set_property;
 
 	esm_class->is_row_selected    = etsm_is_row_selected    ;
 	esm_class->foreach            = etsm_foreach            ;
 	esm_class->clear              = etsm_clear              ;
-#if 0
 	esm_class->selected_count     = etsm_selected_count     ;
-#endif
 	esm_class->select_all         = etsm_select_all         ;
 	esm_class->invert_selection   = etsm_invert_selection   ;
 	esm_class->row_count          = etsm_row_count          ;
@@ -1220,23 +1379,47 @@ e_tree_selection_model_class_init (ETreeSelectionModelClass *klass)
 	esm_class->move_selection_end = etsm_move_selection_end ;
 	esm_class->set_selection_end  = etsm_set_selection_end  ;
 
-	gtk_object_add_arg_type ("ETreeSelectionModel::cursor_row", GTK_TYPE_INT,
-				 GTK_ARG_READWRITE, ARG_CURSOR_ROW);
-	gtk_object_add_arg_type ("ETreeSelectionModel::cursor_col", GTK_TYPE_INT,
-				 GTK_ARG_READWRITE, ARG_CURSOR_COL);
-	gtk_object_add_arg_type ("ETreeSelectionModel::model", E_TREE_MODEL_TYPE,
-				 GTK_ARG_READWRITE, ARG_MODEL);
-	gtk_object_add_arg_type ("ETreeSelectionModel::etta", E_TREE_TABLE_ADAPTER_TYPE,
-				 GTK_ARG_READWRITE, ARG_ETTA);
-	gtk_object_add_arg_type ("ETreeSelectionModel::ets", E_TREE_SORTED_TYPE,
-				 GTK_ARG_READWRITE, ARG_ETS);
+	g_object_class_install_property (object_class, PROP_CURSOR_ROW, 
+					 g_param_spec_int ("cursor_row",
+							   _("Cursor Row"),
+							   /*_( */"XXX blurb" /*)*/,
+							   0, G_MAXINT, 0,
+							   G_PARAM_READWRITE));
+
+	g_object_class_install_property (object_class, PROP_CURSOR_COL, 
+					 g_param_spec_int ("cursor_col",
+							   _("Cursor Column"),
+							   /*_( */"XXX blurb" /*)*/,
+							   0, G_MAXINT, 0,
+							   G_PARAM_READWRITE));
+
+	g_object_class_install_property (object_class, PROP_MODEL, 
+					 g_param_spec_object ("model",
+							      _("Model"),
+							      /*_( */"XXX blurb" /*)*/,
+							      E_TREE_MODEL_TYPE,
+							      G_PARAM_READWRITE));
+
+	g_object_class_install_property (object_class, PROP_ETTA, 
+					 g_param_spec_object ("etta",
+							      _("ETTA"),
+							      /*_( */"XXX blurb" /*)*/,
+							      E_TREE_TABLE_ADAPTER_TYPE,
+							      G_PARAM_READWRITE));
+
+	g_object_class_install_property (object_class, PROP_ETS, 
+					 g_param_spec_object ("ets",
+							      _("ETS"),
+							      /*_( */"XXX blurb" /*)*/,
+							      E_TREE_SORTED_TYPE,
+							      G_PARAM_READWRITE));
 }
 
 ESelectionModel *
 e_tree_selection_model_new (void)
 {
-	return gtk_type_new(e_tree_selection_model_get_type());
+	return g_object_new (E_TREE_SELECTION_MODEL_TYPE, NULL);
 }
 
 E_MAKE_TYPE(e_tree_selection_model, "ETreeSelectionModel", ETreeSelectionModel,
-	    e_tree_selection_model_class_init, e_tree_selection_model_init, PARENT_TYPE);
+	    e_tree_selection_model_class_init, e_tree_selection_model_init, PARENT_TYPE)
